/***************************************************************************
                          hbci.cpp  -  description
                             -------------------
    begin                : Sun May 27 2001
    copyright            : (C) 2001 by fabian kaiser
    email                : fabian.kaiser@gmx.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <stdio.h>
#include <list>

#include "error.h"
#include "pointer.h"
#include "auth.h"
#include "interactor.h"
#include "hbci.h"

namespace HBCI {


int Hbci::_debugLevel=0;

int Hbci::debugLevel() {
    return _debugLevel;
}


void Hbci::setDebugLevel(int l) {
    _debugLevel=l;
}


Hbci::Hbci(bool readonly, bool retrievalonly){
    _readonly=readonly;
    _retrievalOnly = retrievalonly;

    // set the system name and version to default values
    _systemName=HBCI_SYSTEM_NAME;
    _systemVersion=OPENHBCI_VERSION_STRING;
    /* this is a default authentificator, this may be overwritten by the
     * main program.*/
    _auther.setDescription("HBCI::authentificator (HBCIAuth)");
    _auther=new Auth(this);
    _auther.setObjectDescription("HBCIAuth");

    _interactor.setDescription("HBCI::interactor");
    _interactor=new Interactor();
    _interactor.setObjectDescription("HBCIInteractor");
}


Hbci::~Hbci(){
}

void Hbci::libraryVersion(int &major,
			  int &minor,
			  int &patchlevel) {
    major=OPENHBCI_VERSION_MAJOR;
    minor=OPENHBCI_VERSION_MINOR;
    patchlevel=OPENHBCI_VERSION_PATCHLEVEL;
};
bool Hbci::hasLibchipcard ()
{
#ifdef HAS_LIBCHIPCARD
    return true;
#else
    return false;
#endif
}
    
} // namespace HBCI

// ----------- C Wrappers

void HBCI_Hbci_setInteractor(HBCI_Hbci *h, HBCI_Interactor *inter, 
			int autoDelete)
{
    assert(h);
    HBCI::Pointer<HBCI_Interactor> ip = inter;
    ip.setAutoDelete(autoDelete); 
    h->setInteractor(ip);
}
HBCI_Interactor *HBCI_Hbci_interactor(const HBCI_Hbci *h)
{
    assert(h);
    return h->interactor().ptr();
}
void HBCI_Hbci_libraryVersion(int *major, int *minor, int *plevel)
{
    assert(major);
    assert(minor);
    assert(plevel);
    HBCI::Hbci::libraryVersion(*major, *minor, *plevel);
}
int HBCI_Hbci_hasLibchipcard()
{
    return HBCI::Hbci::hasLibchipcard();
}
const char *HBCI_Hbci_systemName(const HBCI_Hbci *h)
{
    assert(h);
    return h->systemName().c_str();
}
void HBCI_Hbci_setSystemName(HBCI_Hbci *h, const char *n)
{
    assert(h);
    h->setSystemName(string(n ? n : ""));
}
const char *HBCI_Hbci_systemVersion(const HBCI_Hbci *h)
{
    assert(h);
    return h->systemVersion().c_str();
}
void HBCI_Hbci_setDebugLevel(int l)
{
    HBCI::Hbci::setDebugLevel(l);
}


