/***************************************************************************
 $RCSfile: mediumrdh.cpp,v $
                             -------------------
    cvs         : $Id: mediumrdh.cpp,v 1.11 2002/10/26 23:28:18 aquamaniac Exp $
    begin       : Sun Nov 18 2001
    copyright   : (C) 2001 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

/*
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h> // creat, write, close
#include <fcntl.h>  // creat

#include <string>
#include <list>
#include "mediumrdh.h"
#include "hbcistring.h"
#include "cryptkey.h"
#include "deskey.h"
#include "rsakey.h"
#include "hbcistring.h"


// global
#define MEDIUMRDH_TAG_SEQ               (unsigned char)0xc1
#define MEDIUMRDH_TAG_USER              (unsigned char)0xc2
#define MEDIUMRDH_TAG_INST              (unsigned char)0xc3
#define MEDIUMRDH_TAG_CRYPT             (unsigned char)0xc4

// userdata
#define MEDIUMRDH_TAG_USER_PUBSIGNKEY    (unsigned char)0xc1
#define MEDIUMRDH_TAG_USER_PRIVSIGNKEY   (unsigned char)0xc2
#define MEDIUMRDH_TAG_USER_PUBCRYPTKEY   (unsigned char)0xc3
#define MEDIUMRDH_TAG_USER_PRIVCRYPTKEY  (unsigned char)0xc4
#define MEDIUMRDH_TAG_USER_ID            (unsigned char)0x05

// instdata
#define MEDIUMRDH_TAG_INST_PUBSIGNKEY    (unsigned char)0xc1
#define MEDIUMRDH_TAG_INST_PUBCRYPTKEY   (unsigned char)0xc2
#define MEDIUMRDH_TAG_INST_COUNTRY       (unsigned char)0x03
#define MEDIUMRDH_TAG_INST_CODE          (unsigned char)0x04
#define MEDIUMRDH_TAG_INST_SYSTEMID      (unsigned char)0x05
#define MEDIUMRDH_TAG_INST_USER          (unsigned char)0x06
#define MEDIUMRDH_TAG_INST_ADDR          (unsigned char)0x07

// keydata
#define MEDIUM_RDH_TAG_KEY_ISPUBLIC (unsigned char) 0x01
#define MEDIUM_RDH_TAG_KEY_ISCRYPT  (unsigned char) 0x02
#define MEDIUM_RDH_TAG_KEY_OWNER    (unsigned char) 0x03
#define MEDIUM_RDH_TAG_KEY_VERSION  (unsigned char) 0x04
#define MEDIUM_RDH_TAG_KEY_NUMBER   (unsigned char) 0x05
#define MEDIUM_RDH_TAG_KEY_MODULUS  (unsigned char) 0x06
#define MEDIUM_RDH_TAG_KEY_EXP      (unsigned char) 0x07
#define MEDIUM_RDH_TAG_KEY_N        (unsigned char) 0x08
#define MEDIUM_RDH_TAG_KEY_P        (unsigned char) 0x09
#define MEDIUM_RDH_TAG_KEY_Q        (unsigned char) 0x0a
#define MEDIUM_RDH_TAG_KEY_DMP1     (unsigned char) 0x0b
#define MEDIUM_RDH_TAG_KEY_DMQ1     (unsigned char) 0x0c
#define MEDIUM_RDH_TAG_KEY_IQMP     (unsigned char) 0x0d
#define MEDIUM_RDH_TAG_KEY_D        (unsigned char) 0x0e


namespace HBCI {

const unsigned int MediumRDH::minPinSize = 4;

MediumRDH::userData::userData() {
    userPubSignatureKey.setDescription("MediumRDH::userData::userPubSignatureKey");
    userPrivateSignatureKey.setDescription("MediumRDH::userData::userPrivateSignatureKey");
    userPubCryptKey.setDescription("MediumRDH::userData::userPubCryptKey");
    userPrivateCryptKey.setDescription("MediumRDH::userPrivateCryptKey");
}


MediumRDH::userData::~userData() {
}


MediumRDH::instData::instData() {
    country=280;
    instPubSignKey.setDescription("MediumRDH::instData::instPubSignKey");
    instPubCryptKey.setDescription("MediumRDH::instData::instPubCryptKey");
}


MediumRDH::instData::~instData() {
}


MediumRDH::MediumRDH(const Hbci *hbci, const string &path)
: Medium(hbci)
,_path(path)
,_currentUser(0)
,_currentInst(0)
,_mountcount(0)
,_seq(1)
,_mounter(0)
{
}



MediumRDH::~MediumRDH() {
}


bool MediumRDH::selectContext(int country,
                                  const string &instcode,
                                  const string &userid){
    userData *ud;
    instData *id;

    // if the context is the same as the current then do nothing
    if (_currentInst && _currentUser)
        if ((_currentInst->instcode==instcode) &&
            (_currentUser->userId==userid))
            return true;

    // mount the medium
    if (!isMounted())
        return false;

    // find institute
    id=_findInst(instcode);
    if (id==0)
        // instcode not found
        return false;
    // find user
    ud=_findUser(userid);
    if (ud==0)
        // userid not found
        return false;
    // does the inst have that user ?
    if (!_hasUser(id,userid))
        // no, return
        return false;

    // ok, we found the context
    _currentInst=id;
    _currentUser=ud;

    if (!_currentInst->instPubSignKey.isValid())
        _hbci->interactor().ref().msgStateResponse("WARNING: No instPubSignkey.\n");
    if (!_currentInst->instPubCryptKey.isValid())
        _hbci->interactor().ref().msgStateResponse("WARNING: No instPubCryptkey.\n");

    return true;
}


Error MediumRDH::mountMedium(Pointer<User> user,
			     const string &pin){
    Error herr;

    if (_mountcount) {
        if (!pin.empty())
            _pin=pin;
        _mountcount++;
	return Error();
    }
    try {
        _mounter = user;
	
        readFile(_path,pin);
    }
    catch (Error err){
        _pin.erase();
	return err;
    }
    _mountcount++;
    if (!pin.empty())
        _pin=pin;
    return Error();
}


void MediumRDH::unmountMedium(const string &pin){
    Error herr;

    _mountcount--;
    if (_mountcount<0)
        _mountcount=0;
    writeFile(_path,pin);
    if (!_mountcount){
        _pin.erase();
        _mounter=0;
    }
}


bool MediumRDH::isMounted() {
    return _mountcount;
}


string MediumRDH::mediumId() const {
    if (!_currentInst)
        return "";
    return _currentInst->systemid;
}


string MediumRDH::createMessageKey() const {
    DESKey desKey;

    desKey=DESKey::createKey();
    return desKey.getKeyString();
}


string MediumRDH::encryptKey(const string &srckey){
    if (!_currentInst)
        throw Error("MediumRDH::encryptKey",
                        "No Context.",0);

    if (!_currentInst->instPubCryptKey.isValid())
        throw Error("MediumRDH::encryptKey",
                        "No Key.",0);
    _currentInst->instPubCryptKey.ref().setData(srckey);
    _currentInst->instPubCryptKey.ref().encrypt();
    return _currentInst->instPubCryptKey.ref().getData();
}


string MediumRDH::decryptKey(const string &srckey){
    string result;

    if (!_currentUser)
        throw Error("MediumRDH::decryptKey",
                        "No Context.",0);

    _currentUser->userPrivateCryptKey.ref().setData(srckey);
    _currentUser->userPrivateCryptKey.ref().decrypt();
    result=_currentUser->userPrivateCryptKey.ref().getData();
    // unpadd
    result =  result.substr(result.length() - 16);
    return result;
}


bool MediumRDH::verify(const string &data, const string &signature){
    if (!_currentInst)
        throw Error("MediumRDH::verify",
                        "No Context.",0);
    _currentInst->instPubSignKey.ref().setData(data);
    return _currentInst->instPubSignKey.ref().verify(signature);
}


string MediumRDH::sign(const string &data){
    if (!_currentUser)
        throw Error("MediumRDH::sign",
                        "No Context.",0);
    _currentUser->userPrivateSignatureKey.ref().setData(data);
    _currentUser->userPrivateSignatureKey.ref().sign();
    return _currentUser->userPrivateSignatureKey.ref().getData();
}


string MediumRDH::encrypt(const string &src, const string &key){
    DESKey desKey;

    desKey.setKeyString(key);
    desKey.setData(src);
    desKey.encrypt();
    return desKey.getData();
}


string MediumRDH::decrypt(const string &src, const string &key){
    DESKey desKey;

    desKey.setKeyString(key);
    desKey.setData(src);
    desKey.decrypt();
    return desKey.getData();
}


unsigned int MediumRDH::nextSEQ(){
    _seq++;
    writeFile(_path,_pin);
    return _seq;
}


int MediumRDH::signKeyNumber() const {
    if (!_currentUser)
        return 0;
    return _currentUser->userPrivateSignatureKey.ref().number();
}


int MediumRDH::signKeyVersion() const {
    if (!_currentUser)
        return 0;
    return _currentUser->userPrivateSignatureKey.ref().version();
}


int MediumRDH::cryptKeyNumber() const {
    if (!_currentInst)
        return 0;
    if (!_currentInst->instPubCryptKey.isValid())
        return 0;
    return _currentInst->instPubCryptKey.ref().number();
}


int MediumRDH::cryptKeyVersion() const {
    if (!_currentInst)
        return 0;
    if (!_currentInst->instPubCryptKey.isValid())
        return 0;
    return _currentInst->instPubCryptKey.ref().version();
}


bool MediumRDH::createUserKeys(){
    RSAKey *k1;
    RSAKey *k2;

    if (!_currentUser)
        return false;

    RSAKey::generateKeyPair(DEFAULT_KEY_LENGTH,
                                &k1,
                                &k2);
    _currentUser->userPrivateSignatureKey=k1;
    _currentUser->userPubSignatureKey=k2;

    _currentUser->userPrivateSignatureKey.setObjectDescription("userPrivateSignatureKey");
    _currentUser->userPrivateSignatureKey.ref().setToSignKey(true);
    _currentUser->userPrivateSignatureKey.ref().setUserId(_currentUser->userId);
    _currentUser->userPubSignatureKey.setObjectDescription("userPubSignatureKey");
    _currentUser->userPubSignatureKey.ref().setToSignKey(true);
    _currentUser->userPubSignatureKey.ref().setUserId(_currentUser->userId);

    RSAKey::generateKeyPair(DEFAULT_KEY_LENGTH,
                                &k1,
                                &k2);
    _currentUser->userPrivateCryptKey=k1;
    _currentUser->userPubCryptKey=k2;
    _currentUser->userPrivateCryptKey.setObjectDescription("userPrivateCryptKey");
    _currentUser->userPrivateCryptKey.ref().setToSignKey(false);
    _currentUser->userPrivateCryptKey.ref().setUserId(_currentUser->userId);
    _currentUser->userPubCryptKey.setObjectDescription("userPubCryptKey");
    _currentUser->userPubCryptKey.ref().setToSignKey(false);
    _currentUser->userPubCryptKey.ref().setUserId(_currentUser->userId);
    writeFile(_path,_pin);
    return true;
}


  /*bool MediumRDH::setInstituteKeys(Pointer<RSAKey> signkey,
                                     Pointer<RSAKey> cryptkey){
    if (!_currentInst)
        return false;


    _currentInst->instPubSignKey=signkey;
    _currentInst->instPubCryptKey=cryptkey;
    writeFile(_path,_pin);
    return true;
    }*/


bool MediumRDH::setInstituteCryptKey(Pointer<RSAKey> cryptkey){
    if (!_currentInst)
        return false;

    if (Hbci::debugLevel() > 1)
	fprintf(stderr, "MediumRDH::setInstituteCryptKey: Setting Inst object %p to instCryptKey %p\n", _currentInst, cryptkey.ptr());
    
    _currentInst->instPubCryptKey=cryptkey;
    writeFile(_path,_pin);
    return true;
}


bool MediumRDH::setInstituteSignKey(Pointer<RSAKey> signkey){
    if (!_currentInst)
        return false;

    _currentInst->instPubSignKey=signkey;
    writeFile(_path,_pin);
    return true;
}


Pointer<RSAKey> MediumRDH::instPubCryptKey() const {
    if (!_currentInst)
        return 0;

    return _currentInst->instPubCryptKey;

}


Pointer<RSAKey> MediumRDH::instPubSignKey() const {
    if (!_currentInst)
        return 0;

    return _currentInst->instPubSignKey;
}

bool MediumRDH::hasInstSignKey() const {
    return instPubSignKey().isValid();
}


Pointer<RSAKey> MediumRDH::userPubCryptKey() const {
    if (!_currentUser)
        return 0;
    return _currentUser->userPubCryptKey;
}


Pointer<RSAKey> MediumRDH::userPubSignKey() const {
    if (!_currentUser) {
        return 0;
    }
    return _currentUser->userPubSignatureKey;
}


void MediumRDH::setSystemId(const string &newid){
    if (_currentInst) {
        _currentInst->systemid=newid;
        writeFile(_path,_pin);
    }
}


MediumRDH::userData *
MediumRDH::_findUser(const string &userid){
    for (list<userData*>::const_iterator iter=lusers.begin();
         iter!=lusers.end(); iter++)
        if ((*iter)->userId==userid)
            return (*iter);
    return 0;
}


bool MediumRDH::_hasUser(const MediumRDH::instData *id, 
			     const string &userid) const {
    if (id==0)
        return false;
    for (list<string>::const_iterator iter=id->users.begin();
         iter!=id->users.end(); iter++)
        if ((*iter)==userid)
            return true;
    return false;
}


MediumRDH::instData *MediumRDH::_findInst(const string &instid){
    for (list<instData*>::const_iterator iter=linst.begin();
         iter!=linst.end(); iter++) {
        if ((*iter)->instcode==instid)
            return (*iter);
    }
    return 0;
}


Pointer<RSAKey> MediumRDH::_readKey(const string &data){
    Pointer<RSAKey> key;
    RSAKey::keyData kd;
    string v;
    string tag;
    unsigned int t;
    unsigned int pos;

    pos=0;
    while(pos<data.length()) {
        tag=String::nextTLV(data,pos);
        v=String::dataTLV(tag);
        t=String::typeTLV(tag);
        switch(t) {
        case MEDIUM_RDH_TAG_KEY_ISPUBLIC:
            if (v=="yes" || v=="YES")
                kd.isPublic=true;
            else
                kd.isPublic=false;
            break;
        case MEDIUM_RDH_TAG_KEY_ISCRYPT:
            if (v=="yes" || v=="YES")
                kd.isCrypt=true;
            else
                kd.isCrypt=false;
            break;
        case MEDIUM_RDH_TAG_KEY_OWNER:
            kd.owner=v;
            break;
        case MEDIUM_RDH_TAG_KEY_VERSION:
            kd.version=atoi(v.c_str());
            break;
        case MEDIUM_RDH_TAG_KEY_NUMBER:
            kd.number=atoi(v.c_str());
            break;
        case MEDIUM_RDH_TAG_KEY_MODULUS:
            kd.modulus=v;
            break;
        case MEDIUM_RDH_TAG_KEY_EXP:
            kd.exponent=atoi(v.c_str());
            break;
        case MEDIUM_RDH_TAG_KEY_N:
            kd.n=v;
            break;

        case MEDIUM_RDH_TAG_KEY_P:
            kd.p=v;
            break;
        case MEDIUM_RDH_TAG_KEY_Q:
            kd.q=v;
            break;
        case MEDIUM_RDH_TAG_KEY_D:
            kd.d=v;
            break;
        case MEDIUM_RDH_TAG_KEY_DMP1:
            kd.dmp1=v;
            break;
        case MEDIUM_RDH_TAG_KEY_DMQ1:
            kd.dmq1=v;
            break;
        case MEDIUM_RDH_TAG_KEY_IQMP:
            kd.iqmp=v;
            break;
        default:
            break;
        } // switch
        pos+=tag.length();;
    } // while
    key=new RSAKey(&kd);
    return key;
}


MediumRDH::userData *MediumRDH::_readUser(const string &data) {
    MediumRDH::userData *ud;
    string v;
    string tag;
    unsigned int t;
    unsigned int pos;

    ud=new userData;

    pos=0;
    while(pos<data.length()) {
        tag=String::nextTLV(data,pos);
        v=String::dataTLV(tag);
        t=String::typeTLV(tag);
        switch(t) {
        case MEDIUMRDH_TAG_USER_PUBSIGNKEY:
            ud->userPubSignatureKey=_readKey(v);
            break;
        case MEDIUMRDH_TAG_USER_PRIVSIGNKEY:
            ud->userPrivateSignatureKey=_readKey(v);
            break;
        case MEDIUMRDH_TAG_USER_PUBCRYPTKEY:
            ud->userPubCryptKey=_readKey(v);
            break;
        case MEDIUMRDH_TAG_USER_PRIVCRYPTKEY:
            ud->userPrivateCryptKey=_readKey(v);
            break;
        case MEDIUMRDH_TAG_USER_ID:
            ud->userId=v;
            break;
        default:
            break;
        } // switch
        pos+=tag.length();;
    } // while
    return ud;
}


MediumRDH::instData *MediumRDH::_readInst(const string &data){
    MediumRDH::instData *id;
    string v;
    string tag;
    unsigned int t;
    unsigned int pos;

    id=new instData;

    pos=0;
    while(pos<data.length()) {
        tag=String::nextTLV(data,pos);
        v=String::dataTLV(tag);
        t=String::typeTLV(tag);
        switch(t) {
        case MEDIUMRDH_TAG_INST_PUBSIGNKEY:
            id->instPubSignKey=_readKey(v);
            break;
        case MEDIUMRDH_TAG_INST_PUBCRYPTKEY:
            id->instPubCryptKey=_readKey(v);
            break;
        case MEDIUMRDH_TAG_INST_COUNTRY:
            id->country=atoi(v.c_str());
            break;
        case MEDIUMRDH_TAG_INST_SYSTEMID:
            id->systemid=v;
            break;
        case MEDIUMRDH_TAG_INST_CODE:
            id->instcode=v;
            break;
        case MEDIUMRDH_TAG_INST_USER:
            id->users.push_back(v);
            break;
        case MEDIUMRDH_TAG_INST_ADDR:
            id->addr=v;
            break;
        default:
            break;
        } // switch
        pos+=tag.length();;
    } // while
    return id;
}


void MediumRDH::readFile(const string &path, const string &pin) {
    Error herr;
    FILE *f;
    string data;
    char buffer[8192];
    int i;
    string tag;
    unsigned int t;
    string v;
    unsigned int pos;
    instData *id;
    userData *ud;
    string tmp;

    _pin=pin;
    while(_pin.length()<minPinSize) {
        herr=_hbci->authentificator().ref().getSecret(_mounter,
                                                      path,
                                                      _pin);
        if (!herr.isOk()) {
            _hbci->interactor().ref().msgStateResponse("Error: "+herr.errorString());
            throw herr;
        }
    } // while

    f=fopen(path.c_str(),"rb");
    if (f==0)
        throw Error("MediumRDH::readFile()",
                        "Could not fopen file.",0);
    while(!feof(f)) {
        i=fread(buffer,1,sizeof(buffer),f);
        if (i<1) {
            fclose(f);
            throw Error("MediumRDH::readFile()",
                            "Could not read file.",0);
        }
        tmp.assign(buffer,i);
        data+=tmp;
    } // while
    if (ferror(f)) {
        fclose(f);
        throw Error("MediumRDH::readFile()",
                        "Error reading file.",0);
    }
    if (fclose(f))
        throw Error("MediumRDH::readFile()",
                        "Could not close file.",0);
    data.assign(buffer,i);

    // check if data is encrypted
    t=String::typeTLV(data);
    if (t==MEDIUMRDH_TAG_CRYPT) {
        // yes, it is, decrypt it
        DESKey dk(_pin);
        dk.setData(String::dataTLV(data));
        dk.decrypt();
        data=dk.getData();
        // check for correct decryption
        pos=0;
        tag=String::nextTLV(data,pos);
        t=String::typeTLV(tag);
        if (t!=MEDIUMRDH_TAG_SEQ)
            throw Error("MediumRDH::readFile()",
                            "Bad pin.",0);
    }

    // backup the current institute
    string lastCurInstId = "";
    if (NULL != _currentInst)
	lastCurInstId = _currentInst->instcode;

    // Clear old user/institute lists
    for (list<userData*>::iterator it = lusers.begin(); 
	 it != lusers.end(); it++)
	delete (*it);
    _currentUser = 0;
    lusers.clear();

    for (list<instData*>::iterator it = linst.begin();
	 it != linst.end(); it++)
	delete (*it);
    _currentInst = 0;
    linst.clear();

    // now parse the data
    pos=0;
    while(pos<data.length()) {
        tag=String::nextTLV(data,pos);
        v=String::dataTLV(tag);
        t=String::typeTLV(tag);
        switch(t) {
        case MEDIUMRDH_TAG_INST:
            id=_readInst(v);
            if (id) {
		if (Hbci::debugLevel() > 1)
		    fprintf(stderr, "MediumRDH::readFile: Just read Inst object %p with instCryptKey %p\n", 
			    id, id->instPubCryptKey.ptr());
                linst.push_back(id);
	    }
            break;
        case MEDIUMRDH_TAG_USER:
            ud=_readUser(v);
            if (ud)
                lusers.push_back(ud);
            break;
        case MEDIUMRDH_TAG_SEQ:
            _seq=atoi(v.c_str());
            break;
        default:
            break;
        } // switch
        pos+=tag.length();
    } // while

    // reload the current institute
    if (!lastCurInstId.empty())
	for (list<instData*>::iterator it = linst.begin(); 
	     it != linst.end(); it++)
	    if ((*it)->instcode == lastCurInstId)
		{
		    if (Hbci::debugLevel()>1)
			printf("MediumRDH::readFile: re-setting cached old _currentInst.\n");
		    _currentInst = (*it);
		}
    
}


string MediumRDH::_writeKey(Pointer<RSAKey> key) const {
    RSAKey::keyData kd;
    string result;
    char buffer[32];

    if (!key.isValid())
        return "";
    if (!key.ref().getKeyData(&kd))
        return "";

    result+=String::newTLV(MEDIUM_RDH_TAG_KEY_ISPUBLIC,
                               (kd.isPublic)?"YES":"NO");
    result+=String::newTLV(MEDIUM_RDH_TAG_KEY_ISCRYPT,
                               (kd.isCrypt)?"YES":"NO");
    if (!kd.owner.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_OWNER,
                                   kd.owner);

    sprintf(buffer,"%d",kd.number);
    result+=String::newTLV(MEDIUM_RDH_TAG_KEY_NUMBER,
                               buffer);
    sprintf(buffer,"%d",kd.version);
    result+=String::newTLV(MEDIUM_RDH_TAG_KEY_VERSION,
                               buffer);
    sprintf(buffer,"%d",kd.exponent);
    result+=String::newTLV(MEDIUM_RDH_TAG_KEY_EXP,
                               buffer);

    if (!kd.modulus.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_MODULUS,
                                   kd.modulus);
    if (!kd.n.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_N,
                                   kd.n);
    if (!kd.p.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_P,
                                   kd.p);
    if (!kd.q.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_Q,
                                   kd.q);
    if (!kd.d.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_D,
                                   kd.d);
    if (!kd.dmp1.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_DMP1,
                                   kd.dmp1);
    if (!kd.dmq1.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_DMQ1,
                                   kd.dmq1);
    if (!kd.iqmp.empty())
        result+=String::newTLV(MEDIUM_RDH_TAG_KEY_IQMP,
                                   kd.iqmp);
    return result;
}


string MediumRDH::_writeUser(userData &ud) const {
    string result;
    string tag;

    tag=_writeKey(ud.userPubSignatureKey);
    if (!tag.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_USER_PUBSIGNKEY,tag);

    tag=_writeKey(ud.userPrivateSignatureKey);
    if (!tag.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_USER_PRIVSIGNKEY,tag);

    tag=_writeKey(ud.userPubCryptKey);
    if (!tag.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_USER_PUBCRYPTKEY,tag);

    tag=_writeKey(ud.userPrivateCryptKey);
    if (!tag.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_USER_PRIVCRYPTKEY,tag);

    if (!ud.userId.empty()) {
        result+=String::newTLV(MEDIUMRDH_TAG_USER_ID,
                                   ud.userId);
    }
    return result;
}


string MediumRDH::_writeInst(instData &id) const {
    string result;
    char buffer[32];
    string tag;

    tag=_writeKey(id.instPubSignKey);
    if (!tag.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_INST_PUBSIGNKEY,tag);
    tag=_writeKey(id.instPubCryptKey);
    if (!tag.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_INST_PUBCRYPTKEY,tag);

    sprintf(buffer,"%d",id.country);
    result+=String::newTLV(MEDIUMRDH_TAG_INST_COUNTRY,
                               buffer);
    if (!id.instcode.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_INST_CODE,
                                   id.instcode);
    if (!id.systemid.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_INST_SYSTEMID,
                                   id.systemid);
    // writer use ids
    for (list<string>::const_iterator iter=id.users.begin();
         iter!=id.users.end(); iter++) {
        if (!(*iter).empty()) {
            result+=String::newTLV(MEDIUMRDH_TAG_INST_USER,
                                       *iter);
        }
    } // for
    if (!id.addr.empty())
        result+=String::newTLV(MEDIUMRDH_TAG_INST_ADDR,
                                   id.addr);

    return result;
}



void MediumRDH::writeFile(const string &path, const string &pin) {
    Error herr;
    char buffer[32];
    string result;
    string tag;
    string tmp;

    if (!pin.empty())
        _pin=pin;
    while(_pin.length()<minPinSize) {
        herr=_hbci->authentificator().ref().getSecret(_mounter,
                                                      path,
                                                      _pin,
                                                      true);
        if (!herr.isOk()) {
            _hbci->interactor().ref().msgStateResponse("Error: "+herr.errorString());
            if (Hbci::debugLevel() > 0)
		fprintf(stderr,"Error: %s\n",
			herr.errorString().c_str());
            throw Error("MediumRDH::unmountMedium()",
                            "no passwd",0);
        }
    } // while

    // write SEQ
    sprintf(buffer,"%d",_seq);
    result+=String::newTLV(MEDIUMRDH_TAG_SEQ,
                               buffer);

    // write institutes
    for (list<instData*>::const_iterator iterI=linst.begin();
         iterI!=linst.end(); iterI++) {
	if (Hbci::debugLevel() > 1)
	    fprintf(stderr, "MediumRDH::writeFile: writing Inst object %p with instCryptKey %p\n", *iterI,
		   (*iterI)->instPubCryptKey.ptr());
        tag=_writeInst(*(*iterI));
        if (!tag.empty())
            result+=String::newTLV(MEDIUMRDH_TAG_INST,tag);
    }

    // write users
    for (list<userData*>::const_iterator iterU=lusers.begin();
         iterU!=lusers.end(); iterU++) {
        tag=_writeUser(*(*iterU));
        if (!tag.empty())
            result+=String::newTLV(MEDIUMRDH_TAG_USER,tag);
    }

    // encrypt data
    DESKey dk(_pin);
    dk.setData(result);
    dk.encrypt();
    result=String::newTLV(MEDIUMRDH_TAG_CRYPT,
                              dk.getData());

    // now write it
    int fd = creat(path.c_str(), 0600);
    if (fd < 0)
        throw Error("MediumRDH::_writeFile()",
                        "Could not open RDH file for writing.",0);
    if ((int)result.length() != write(fd, result.data(), result.length())){
        close(fd);
        throw Error("MediumRDH::_writeFile()",
                        "Error while writing to RDH file.",0);
    }
    if (0 != close(fd))
        throw Error("MediumRDH::_writeFile()",
        "Could not close file.",0);
}


bool MediumRDH::createContext(int country,
                                  const string &instcode,
                                  const string &userid,
				  const string &pin){
    instData *id;
    userData *ud;

    // find institute
    id=_findInst(instcode);
    // found ?
    if (id==0) {
        // no, create it
        id=new instData;
        id->instcode=instcode;
        id->country=country;
        linst.push_back(id);
    }
    // is the user already part of the inst ?
    if (_hasUser(id,userid))
        // yes, finished
        return true;
    // does the user exist ?
    ud=_findUser(userid);
    if (ud==0) {
        // no, create it
        ud=new userData;
        ud->userId=userid;
        lusers.push_back(ud);
    }
    // add user to institute
    id->users.push_back(userid);
    // write new state

    writeFile(_path,pin);

    // finished
    return true;
}



bool MediumRDH::getContext(int num,
                               int &countrycode,
                               string &instcode,
                               string &userid,
                               string &server) const {
    if (num<1)
        return false;
    for (list<instData*>::const_iterator iter=linst.begin();
         iter!=linst.end(); iter++) {
        if (--num==0) {
            countrycode=(*iter)->country;
            instcode=(*iter)->instcode;
            userid=(*iter)->users.front();
            server=(*iter)->addr;
            return true;
        }
    } // for
    return false;
}



string MediumRDH::getInstIniLetterModulus(bool cryptkey) const {
    if (!cryptkey)
        return instPubSignKey().ref().getIniLetterModulus();
    else
        return instPubCryptKey().ref().getIniLetterModulus();
}
	
string MediumRDH::getInstIniLetterExponent(bool cryptkey) const {
    if (!cryptkey)
        return instPubSignKey().ref().getIniLetterExponent();
    else
        return instPubCryptKey().ref().getIniLetterExponent();
}

string MediumRDH::getInstIniLetterHash(bool cryptkey) const {
    if (!cryptkey)
        return instPubSignKey().ref().getIniLetterHash();
    else
        return instPubCryptKey().ref().getIniLetterHash();
}
int MediumRDH::getInstKeyNumber(bool cryptkey) const{
    return (cryptkey ? 
	    instPubCryptKey().ref().number() :
	    instPubSignKey().ref().number());
}
int MediumRDH::getInstKeyVersion(bool cryptkey) const{
    return (cryptkey ? 
	    instPubCryptKey().ref().version() :
	    instPubSignKey().ref().version());
}


string MediumRDH::getUserIniLetterModulus() const {
    return userPubSignKey().ref().getIniLetterModulus();
}


string MediumRDH::getUserIniLetterExponent() const {
    return userPubSignKey().ref().getIniLetterExponent();
}


string MediumRDH::getUserIniLetterHash() const {
    return userPubSignKey().ref().getIniLetterHash();
}


int MediumRDH::getUserKeyNumber() const{
    return userPubCryptKey().ref().number();
}


int MediumRDH::getUserKeyVersion() const{
    return userPubCryptKey().ref().version();
}

} // namespace HBCI

HBCI_MediumRDH *HBCI_Medium_MediumRDH(HBCI_Medium *h)
{
    return dynamic_cast<HBCI_MediumRDH*>(h);
}
int HBCI_MediumRDH_hasInstSignKey(const HBCI_MediumRDH *h)
{
    assert(h);
    return (h->hasInstSignKey());
}
char *
HBCI_MediumRDH_getInstIniLetterExponent(const HBCI_MediumRDH *h, 
					int usecrypt)
{
    assert(h);
    return hbci_strdup(h->getInstIniLetterExponent(usecrypt));
}
char *
HBCI_MediumRDH_getInstIniLetterHash(const HBCI_MediumRDH *h, 
				    int usecrypt)
{
    assert(h);
    return hbci_strdup(h->getInstIniLetterHash(usecrypt));
}
char *
HBCI_MediumRDH_getInstIniLetterModulus(const HBCI_MediumRDH *h, 
				       int usecrypt)
{
    assert(h);
    return hbci_strdup(h->getInstIniLetterModulus(usecrypt));
}
int HBCI_MediumRDH_getInstKeyNumber(const HBCI_MediumRDH *h, 
					       int usecrypt)
{
    assert(h);
    return h->getInstKeyNumber(usecrypt);
}
int HBCI_MediumRDH_getInstKeyVersion(const HBCI_MediumRDH *h, 
						int usecrypt)
{
    assert(h);
    return h->getInstKeyVersion(usecrypt);
}
char *
HBCI_MediumRDH_getUserIniLetterExponent(const HBCI_MediumRDH *h)
{
    assert(h);
    return hbci_strdup(h->getUserIniLetterExponent());
}
char *
HBCI_MediumRDH_getUserIniLetterHash(const HBCI_MediumRDH *h)
{
    assert(h);
    return hbci_strdup(h->getUserIniLetterHash());
}
char *
HBCI_MediumRDH_getUserIniLetterModulus(const HBCI_MediumRDH *h)
{
    assert(h);
    return hbci_strdup(h->getUserIniLetterModulus());
}
int HBCI_MediumRDH_getUserKeyNumber(const HBCI_MediumRDH *h)
{
    assert(h);
    return h->getUserKeyNumber();
}
int HBCI_MediumRDH_getUserKeyVersion(const HBCI_MediumRDH *h)
{
    assert(h);
    return h->getUserKeyVersion();
}

