/* test-value.cpp
 * Test-drive the implementation of HBCI::Value
 * by Christian Stimming <stimming@tuhh.de>
 *
 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>

#include <openhbci/hbci.h>
#include <openhbci/value.h>

#include <iomanip>
#include <stdexcept>
#include <locale.h>

using namespace HBCI;

const char* values[] = {
    "12,34",
    "-12,34",
    "0,12",
    "-0,12",
    // Test values from HBCI-Spec II.5.1
    "100,00",
    "100,20",
    "4567,89",
    "0",
    "0,50",
    // Values with currency
    "12,34:EUR",
    "-12,34:EUR",
    "12,345:EUR",
    "-12,345:EUR",
    // precision== 0 here, so the nachkommastellen will be cut off
    "12,34:JPY", 
    "-12,34:JPY",
    // This would have given a buffer overflow with the old
    // implementation. Due to limited range in computers, this will be
    // rounded at 10^17 nevertheless.
    "12345678901234567,89:EUR",
    // Now some errornous values:
    // Two ':'
    "-12,34:EUR:EUR",
    // currency with length other than 3, but will silently be
    // truncated
    "-12,34:Euro",
    // mixed point and comma
    "-1.234,00:EUR",
    // several comma
    "-1,234,00:EUR",
    // several points
    "-1.234.00:EUR",
    // wrong number format, but only first part is interpreted
    "1 234,00:EUR",
    // those will silently be accepted
    "1.2e3:EUR",
    0l
};

int main(int argc, char **argv) {
    const char *v;
    const char **vptr;
    //unsigned int maxwidth = 0;

    HBCI::Hbci::setDebugLevel(3);
    setlocale(LC_ALL, "");
    
    //for (vptr = values; v=*vptr; vptr++) 
    //maxwidth = max(strlen(v), maxwidth);

    for (vptr = values; (v=*vptr); vptr++) {
	try {
	    HBCI::Value hv(v);
	    cout << "Orig.string -> Value::toString(): '" 
		 << v << "' -> '" 
		//<< hv.getValue() << "', '"
		//<< hv.getCurrency() << "' -> '"
		 << hv.toString() 
		//<< "' or '" << hv.toReadableString() 
		 << "'" << endl;
	    HBCI::Value hv2(hv.toString());
	    if (hv.toString() != hv2.toString())
		cout << "  CRITICAL: Second string->Value conversion differs. '"
		     << hv.toString() << "' != '"
		     << hv2.toString() << "'" << endl;
	} catch (HBCI::Error &e) {
	    cout << "At string '" << v 
		 << "' caught Error " << e.message() << endl;
	} catch (out_of_range &o) {
	    cout << "CRITICAL: At string '" << v 
		 << "' caught out_of_range exeption." << endl;
	} catch (...) {
	    cout << "CRITICAL: At string '" << v 
		 << "' caught unknown exception." << endl;
	}
    }

    cout << endl
	 << "  Did you get any CRITICAL errors? If yes, please send a " << endl
	 << "  bugreport to Christian <stimming@tuhh.de>. Thanks for testing." << endl;
    return 0;	    
}




