/***************************************************************************
 $RCSfile: rdhconverter.cpp,v $
                             -------------------
    cvs         : $Id: rdhconverter.cpp,v 1.4 2003/01/28 17:19:30 cstim Exp $
    begin       : Wed Aug 28 2002
    copyright   : (C) 2002 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#include "rdhconverter.h"
#include <openhbci/file.h>

using namespace std;
namespace HBCI {

RDHConverter::RDHConverter(Pointer<MediumRDH> mold,
                           int idx)
:_mold(mold)
,_idx(idx)
{
    _mnew=new MediumKeyfile(mold.ref().hbci(),
                            mold.ref().mediumName()+".new");
}


RDHConverter::~RDHConverter(){
}


bool RDHConverter::_convertMedium() {
    MediumRDH::instData id;
    MediumRDH::userData ud;
    unsigned int seq;

    if (Hbci::debugLevel()>2)
        cerr<<"MediumKeyfileBase::createMedium\n";

    if (!_mold.ref()._getContext(_idx,id,ud,seq)) {
        if (Hbci::debugLevel()>0)
            cerr<<"Could not get context.\n";
        return false;
    }

    _mnew.ref()._clearMedium();

    // copy data from context
    _mnew.ref()._seq=seq;
    _mnew.ref()._userPubSignatureKey=ud.userPubSignatureKey;
    _mnew.ref()._userPrivateSignatureKey=ud.userPrivateSignatureKey;
    _mnew.ref()._userPubCryptKey=ud.userPubCryptKey;
    _mnew.ref()._userPrivateCryptKey=ud.userPrivateCryptKey;
    _mnew.ref()._userId=ud.userId;

    _mnew.ref()._country=id.country;
    _mnew.ref()._instcode=id.instcode;
    _mnew.ref()._systemid=id.systemid;
    _mnew.ref()._instPubSignKey=id.instPubSignKey;
    _mnew.ref()._instPubCryptKey=id.instPubCryptKey;

    return true;
}


bool RDHConverter::convertMedium(const string &pinold,
                                 const string &pinnew) {
    int countrycode;
    string instcode;
    string userid;
    string server;
    bool b;
    Error err;

    // mount old medium
    err=_mold.ref().mountMedium(0,pinold);
    if (!err.isOk()){
	cerr<<"Could not mount old medium."<<err.errorString()<<"\n";
	return false;
    }

    // get context
    if (!_mold.ref().getContext(_idx,
                                countrycode,
                                instcode,
                                userid,
                                server)) {
        cerr<<"Unable to get context.\n";
        return false;
    }

    // create new medium
    if (!_mnew.ref().createMedium(countrycode,
                                  instcode,
                                  userid,
                                  pinnew)) {
        cerr<<"Could not create new medium.\n";
        return false;
    }

    // mount new medium
    err=_mnew.ref().mountMedium(0,pinnew);
    if (!err.isOk()) {
	cerr<<"Could not mount new medium: "<<err.errorString()<<"\n";
        return false;
    }

    // really convert
    b=_convertMedium();
    _mnew.ref().unmountMedium(pinnew);
    _mold.ref().unmountMedium(pinold);

    if (!b) {
        cerr<<"Could not convert medium.\n";
        return false;
    }

    // rename old file
    File fold(_mold.ref().mediumName());
    err=fold.renameFile(_mold.ref().mediumName()+".old");
    if (!err.isOk()) {
        cerr<<"Could not rename old file: "<<err.errorString()<<"\n";
        return false;
    }

    // rename new file (to name of old medium!)
    File fnew(_mnew.ref().mediumName());
    err=fnew.renameFile(_mold.ref().mediumName());
    if (!err.isOk()) {
        cerr<<"Could not rename new file: "<<err.errorString()<<"\n";
        return false;
    }
    return true;
}


} // namespace


